from ..utils.helper_functions import _parse_path_variables
from ..utils.constants import DEFAULT_LAYOUT_TOKEN

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Awaitable,
    Optional,
    ClassVar,
    Tuple,
)
from dash.development.base_component import Component
from dash._utils import AttributeDict
from pydantic import BaseModel, Field
from uuid import UUID


class RouteConfig(BaseModel):
    default_child: str | None = None
    is_static: bool | None = None
    title: str | None = None
    description: str | None = None
    name: str | None = None
    order: int | None = None
    image: str | None = None
    image_url: str | None = None
    redirect_from: List[str] | None = None


class RouterResponse(BaseModel):
    response: Dict[str, any]
    mimetype: str = "application/json"
    multi: bool = False

    class Config:
        arbitrary_types_allowed = True


class PageNode(BaseModel):
    segment_value: str = Field(alias="_segment")  # Changed to use alias
    node_id: str
    layout: Callable[..., Awaitable[Component]] | Component
    module: str
    parent_id: Optional[str] = None
    path: Optional[str] = None
    is_static: bool = False
    is_root: Optional[bool] = None
    default_child: Optional[str] = None
    child_nodes: Dict[str, UUID] = Field(default_factory=dict)
    slots: Dict[str, str] = Field(default_factory=dict)
    path_template: Optional[str] = None
    loading: Optional[Callable[..., Awaitable[Component]] | Component] = None
    error: Optional[Callable[..., Awaitable[Component]] | Component] = None
    endpoint: Optional[Callable[..., Awaitable[Any]]] = None
    endpoint_inputs: Optional[List[Any]] = None

    class Config:
        arbitrary_types_allowed = True
        populate_by_name = True

    @property
    def is_slot(self) -> bool:
        segment = self.segment_value
        return segment.startswith("(") and segment.endswith(")")

    @property
    def is_path_template(self) -> bool:
        segment = self.segment_value
        return segment.startswith("[") and segment.endswith("]")

    @property
    def segment(self) -> str:
        if self.is_path_template:
            return self.segment_value.strip("[]")

        if self.is_slot:
            return self.segment_value.strip("()")

        formatted_segment = self.segment_value.replace("_", "-").replace(" ", "-")
        return formatted_segment

    def register_slot(self, node: "PageNode") -> None:
        if node.segment in self.slots:
            raise KeyError(f"{node.segment} is already registered as slot!")

        self.slots[node.segment] = node.node_id

    def register_route(self, node: "PageNode") -> None:
        if node.segment in self.slots:
            raise KeyError(f"{node.segment} is already registered as parallel route!")

        self.child_nodes[node.segment] = node.node_id

    def register_path_template(self, node: "PageNode") -> None:
        if self.path_template:
            raise ValueError(f"{node.segment} already has a path template!")

        self.path_template = node.node_id

    def create_segment_key(self, var):
        if not self.is_path_template:
            return self.segment_value

        path_key = self.segment_value
        path_var = var or DEFAULT_LAYOUT_TOKEN
        filled_template = path_key.replace(self.segment, path_var)
        path_template_key = path_key + filled_template
        return path_template_key

    def get_child_node(self, segment: str) -> Optional["PageNode"]:
        if self.default_child and not segment:
            default_node_id = self.child_nodes.get(self.default_child)
            return RouteTable.get_node(default_node_id)

        # Try to match a parallel child first
        if child_id := self.child_nodes.get(segment):
            return RouteTable.get_node(child_id)

        # Otherwise, check the slots for a node with a path template
        if self.path_template:
            return RouteTable.get_node(self.path_template)

    def get_slots(self):
        return {key: RouteTable.get_node(val) for key, val in self.slots.items()}


class RouteTable:

    _table: ClassVar[Dict[str, PageNode]] = {}

    def __new__(cls):
        raise TypeError("RouteTable is a static class and should not be instantiated")

    @classmethod
    def add_node(cls, node: PageNode) -> None:
        if node.node_id in cls._table:
            raise KeyError(f"{node.segment} is already registered!")

        cls._table[node.node_id] = node

    @classmethod
    def get_node(cls, node_id: str) -> PageNode:
        return cls._table.get(node_id)


class RouteTree:

    _static_routes: ClassVar[Dict[str, str]] = {}
    _dynamic_routes: ClassVar[AttributeDict] = AttributeDict(
        routes={}, path_template=None
    )

    def __new__(cls):
        raise TypeError("RouteTree is a static class and should not be instantiated")

    @classmethod
    def add_static_route(cls, new_node: PageNode) -> None:
        if new_node.path in cls._static_routes:
            raise KeyError(
                f"{new_node.segment} with path {new_node.path} is already present in static routes"
            )

        cls._static_routes[new_node.path] = new_node

    @classmethod
    def add_root_route(cls, new_node: PageNode) -> None:
        if new_node.is_path_template:
            if cls._dynamic_routes.path_template:
                raise ValueError(f"{new_node.segment} already has a path template!")

            cls._dynamic_routes.path_template = new_node.node_id

        if new_node.segment in cls._dynamic_routes.routes:
            raise KeyError(
                f"{new_node.segment} with path {new_node.path} is already present in static routes"
            )

        cls._dynamic_routes.routes[new_node.segment] = new_node.node_id

    @classmethod
    def get_root_node(cls, ctx) -> Optional[PageNode]:
        missed_segments: str = None
        node: PageNode = None

        for segment in ctx.segments:
            if missed_segments:
                segment = missed_segments + "/" + segment

            if node_id := cls._dynamic_routes.routes.get(segment):
                node = RouteTable.get_node(node_id)
                ctx.segments.pop(0) if ctx.segments else None
                return node

            if node_id := cls._dynamic_routes.path_template:
                node = RouteTable.get_node(node_id)
                ctx.path_vars[node.segment] = segment
                return node

            missed_segments = segment
            ctx.segments.pop(0) if ctx.segments else None

        return node

    @classmethod
    def get_active_root_node(cls, ctx, ignore_empty_folders: bool):
        active_node = cls.get_root_node(ctx)
        ctx.segments = list(reversed(ctx.segments))

        while ctx.segments:

            if active_node is None:
                return active_node

            next_segment = ctx.peek_segment()
            segment_key = active_node.create_segment_key(next_segment)
            segment_loading_state = ctx.get_node_state(segment_key)

            if not segment_loading_state:
                return active_node

            if active_node.is_path_template:

                if len(ctx.segments) <= 1:
                    return active_node

                ctx.consume_path_var(active_node)
                next_segment = ctx.peek_segment()

            ctx.set_node_state(active_node, "done", segment_key)
            ctx.set_silent_loading_states(active_node, "done")

            if child_node := active_node.get_child_node(next_segment):
                if not child_node.is_path_template:
                    ctx.pop_segment()

                active_node = child_node
                continue

            ctx.merge_segments(ignore_empty_folders)

        return active_node

    @classmethod
    def add_node(cls, new_node: PageNode, parent_node: PageNode | None) -> None:
        if new_node.is_static:
            cls.add_static_route(new_node)
            return

        if new_node.is_root:
            cls.add_root_route(new_node)
            return

        if new_node.is_slot:
            parent_node.register_slot(new_node)
            return

        if new_node.is_path_template:
            parent_node.register_path_template(new_node)
            return

        parent_node.register_route(new_node)

    @classmethod
    def get_static_route(cls, ctx) -> Tuple[PageNode, Dict[str, Any]]:
        path_variables = {}
        if not ctx.pathname:
            index_node = cls._static_routes.get("/")
            return index_node, path_variables

        for page_path, page_id in cls._static_routes.items():
            if "[" in page_path and "]" in page_path:
                path_variables = _parse_path_variables(ctx.pathname, page_path)
                if path_variables:
                    page_node = RouteTable.get_node(page_id)
                    return page_node, path_variables

            if ctx.pathname == page_path:
                page_node = RouteTable.get_node(page_id)
                return page_node, path_variables

        return None, path_variables



from ast import Call
import asyncio
from dataclasses import dataclass, field
from functools import partial
from typing import Any, Callable, Dict, List, Literal, Optional
from uuid import UUID

from ..utils.constants import DEFAULT_LAYOUT_TOKEN, REST_TOKEN
from .routing import PageNode, RouteTable
from pydantic import BaseModel


class LoadingState(BaseModel):
    state: Literal["lacy", "done", "hidden"]
    node_id: str
    updated: bool = False

    def update_state(self, state: Literal["lacy", "done", "hidden"]) -> None:
        self.state = state
        self.updated = True


@dataclass
class RoutingContext:
    """Encapsulates all routing state for a single request"""

    pathname: str
    query_params: Dict[str, Any]
    resolve_type: Literal["search", "url", "lacy"]
    path_vars: Dict[str, str] = field(default_factory=dict)
    endpoints: Dict[UUID, Callable] = field(default_factory=dict)
    segments: List[str] = field(default_factory=list)
    _loading_states: Dict[str, Any] = field(default_factory=dict, repr=False)

    @property
    def variables(self):
        return {**self.query_params, **self.path_vars}

    @classmethod
    def from_request(
        cls,
        pathname: str,
        query_params: Dict[str, Any],
        loading_state_dict: Dict[str, Dict],
        resolve_type: Literal["search", "url", "lacy"],
    ) -> "RoutingContext":
        """Create context from request data"""
        path = pathname.strip("/")
        segments = [seg for seg in path.split("/") if seg] if path else []
        # Build _loading_states dict
        _loading_states = {
            segment_key: LoadingState(**ils)
            for segment_key, ils in loading_state_dict.items()
        }
        return cls(
            pathname=pathname,
            query_params=query_params,
            segments=segments,
            resolve_type=resolve_type,
            _loading_states=_loading_states,
        )

    def get_node_state(self, segment_key: Optional[str] = None) -> Optional[str]:
        """Get loading state for a node"""
        ls = self._loading_states.get(segment_key)
        if ls:
            return ls.state
        return None

    def set_node_state(
        self, node: PageNode, state: str, segment_key: Optional[str] = None
    ):
        """Set loading state for a node"""
        if segment_key not in self._loading_states:
            self._loading_states[segment_key] = LoadingState(
                state=state, node_id=node.node_id, updated=True
            )
        else:
            self._loading_states[segment_key].update_state(state)

    def add_endpoint(self, node: PageNode):
        partial_endpoint = partial(node.endpoint, **self.variables)
        self.endpoints[node.node_id] = partial_endpoint

    def should_lazy_load(
        self, node: PageNode, segment_key: Optional[str] = None
    ) -> bool:
        """Check if node should be lazy loaded"""
        return (
            node.loading is not None
            and self.resolve_type != "lacy"
            and DEFAULT_LAYOUT_TOKEN not in segment_key
        )

    def pop_segment(self) -> Optional[str]:
        """Remove and return the last segment"""
        return self.segments.pop() if self.segments else None

    def peek_segment(self) -> Optional[str]:
        """Peek at the last segment without removing"""
        return self.segments[-1] if self.segments else None

    def consume_path_var(self, node: PageNode) -> Optional[str]:
        """Consume a segment as path variable"""
        if not node.is_path_template:
            return None

        if node.segment == REST_TOKEN:
            rest_value = list(reversed(self.segments))
            self.segments = []
            self.path_vars["rest"] = rest_value
            return rest_value
        else:
            value = self.pop_segment()
            if value:
                self.path_vars[node.segment] = value
            return value

    def merge_segments(self, ignore_empty_folders: bool):
        """Merge segments if empty folder should be ignored"""
        if ignore_empty_folders or len(self.segments) < 2:
            self.pop_segment()

        first = self.segments.pop()
        second = self.segments.pop()
        combined = f"{first}/{second}"
        self.segments.append(combined)

    async def gather_endpoints(self):
        if not self.endpoints:
            return {}

        keys = list(self.endpoints.keys())
        funcs = list(self.endpoints.values())
        results = await asyncio.gather(
            *[func() for func in funcs], return_exceptions=True
        )
        return dict(zip(keys, results))

    def to_loading_state_dict(self) -> Dict[str, Any]:
        """Convert context back to loading state dict for response"""
        return {**self.get_updated_loading_state(), "query_params": self.query_params}

    def get_updated_loading_state(self) -> Dict[str, Dict]:
        """Return only updated loading states as a dict."""
        return {
            key: {"state": state.state, "node_id": state.node_id}
            for key, state in self._loading_states.items()
            if state.updated == True
        }

    def set_silent_loading_states(self, node: PageNode, state: str = "done"):
        """Mark all descendant slots as done"""
        for slot_name, slot_id in node.slots.items():
            slot_node = RouteTable.get_node(slot_id)
            if slot_node:
                self.set_node_state(slot_node, state, slot_name)
                self.set_silent_loading_states(slot_node, state)



from asyncio import iscoroutinefunction

import os
from _plotly_utils.optional_imports import get_module
from plotly.io._json import clean_to_json_compatible, config, JsonConfig
from fnmatch import fnmatch
import re


async def _invoke_layout(func, *func_args, **func_kwargs):
    if iscoroutinefunction(func):
        return await func(*func_args, **func_kwargs)

    if callable(func):
        return func(*func_args, **func_kwargs)

    return func


def recursive_to_plotly_json(component):
    """
    Recursively convert a component to a JSON-serializable structure.
    Handles Plotly components, numpy arrays, pandas objects, dates/times, and other special types.

    Parameters:
    -----------
    component: Any
        The component to convert

    Returns:
    --------
    A JSON-serializable representation of the component
    """
    # Base case: simple types don't need conversion
    if component is None or isinstance(component, (str, int, float, bool)):
        return component

    # Try to handle numpy arrays first
    try:
        import numpy as np

        if isinstance(component, np.ndarray):
            return component.tolist()
        elif np.isscalar(component) and not isinstance(
            component, (bool, int, float, complex)
        ):
            return component.item()
    except (ImportError, AttributeError):
        pass

    # Handle pandas objects
    try:
        import pandas as pd

        if isinstance(component, (pd.Series, pd.DataFrame)):
            return component.to_dict()
        elif isinstance(component, pd.Timestamp):
            return component.isoformat()
        elif component is pd.NaT:
            return None
    except (ImportError, AttributeError):
        pass

    # Handle datetime objects
    try:
        import datetime

        if isinstance(component, (datetime.date, datetime.datetime)):
            return component.isoformat()
    except (ImportError, AttributeError):
        pass

    # Handle decimal
    try:
        import decimal

        if isinstance(component, decimal.Decimal):
            return float(component)
    except (ImportError, AttributeError):
        pass

    # Convert component to plotly json if it has the method
    if hasattr(component, "to_plotly_json"):
        component = component.to_plotly_json()

    # Also try other common serialization methods
    if hasattr(component, "tolist"):
        try:
            return component.tolist()
        except Exception:
            pass

    if hasattr(component, "to_dict"):
        try:
            return component.to_dict()
        except Exception:
            pass

    # Make sure component is a dictionary before checking for "props"
    if isinstance(component, dict):
        # Process props
        for key, value in list(component.items()):
            if isinstance(value, list):
                # Process lists of items
                component[key] = [recursive_to_plotly_json(item) for item in value]
            else:
                # Process single items
                component[key] = recursive_to_plotly_json(value)

    # Handle list-type components
    elif isinstance(component, list):
        component = [recursive_to_plotly_json(item) for item in component]

    # As a last resort, try string representation
    else:
        try:
            return str(component)
        except Exception:
            return None

    return component


def format_relative_path(path: str):
    return path.replace(".", "/").replace("_", "-").replace(" ", "-")


def path_to_module(current_dir: str, module: str):
    module_path = os.path.join(current_dir, module)
    module_path_parts = os.path.splitext(module_path)[0].split(os.sep)
    module_name = ".".join(module_path_parts)
    return module_name


def to_json_plotly(plotly_object, pretty=False, engine=None):
    """
    Convert a plotly/Dash object to a JSON string representation

    Parameters
    ----------
    plotly_object:
        A plotly/Dash object represented as a dict, graph_object, or Dash component

    pretty: bool (default False)
        True if JSON representation should be pretty-printed, False if
        representation should be as compact as possible.

    engine: str (default None)
        The JSON encoding engine to use. One of:
          - "json" for an engine based on the built-in Python json module
          - "orjson" for a faster engine that requires the orjson package
          - "auto" for the "orjson" engine if available, otherwise "json"
        If not specified, the default engine is set to the current value of
        plotly.io.json.config.default_engine.

    Returns
    -------
    str
        Representation of input object as a JSON string

    See Also
    --------
    to_json : Convert a plotly Figure to JSON with validation
    """
    orjson = get_module("orjson", should_load=True)

    # Determine json engine
    if engine is None:
        engine = config.default_engine

    if engine == "auto":
        if orjson is not None:
            engine = "orjson"
        else:
            engine = "json"
    elif engine not in ["orjson", "json"]:
        raise ValueError("Invalid json engine: %s" % engine)

    modules = {
        "sage_all": get_module("sage.all", should_load=False),
        "np": get_module("numpy", should_load=False),
        "pd": get_module("pandas", should_load=False),
        "image": get_module("PIL.Image", should_load=False),
    }

    # Dump to a JSON string and return
    # --------------------------------
    if engine == "json":
        opts = {}
        if pretty:
            opts["indent"] = 2
        else:
            # Remove all whitespace
            opts["separators"] = (",", ":")

        from _plotly_utils.utils import PlotlyJSONEncoder

        return _plotly_object

    elif engine == "orjson":
        JsonConfig.validate_orjson()
        opts = orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY

        if pretty:
            opts |= orjson.OPT_INDENT_2

        # Plotly
        try:
            plotly_object = plotly_object.to_plotly_json()
        except AttributeError:
            pass

        # Try without cleaning
        try:
            return plotly_object

        except TypeError:
            pass

        cleaned = clean_to_json_compatible(
            plotly_object,
            numpy_allowed=True,
            datetime_allowed=True,
            modules=modules,
        )
        return cleaned


def _parse_path_variables(pathname, path_template):
    """
    creates the dict of path variables passed to the layout
    e.g. path_template= "/asset/<asset_id>"
         if pathname provided by the browser is "/assets/a100"
         returns **{"asset_id": "a100"}
    """

    # parse variable definitions e.g. <var_name> from template
    # and create pattern to match
    wildcard_pattern = re.sub("[.*?]", "*", path_template)
    var_pattern = re.sub("[.*?]", "(.*)", path_template)

    # check that static sections of the pathname match the template
    if not fnmatch(pathname, wildcard_pattern):
        return None

    # parse variable names e.g. var_name from template
    var_names = re.findall("[(.*?)]", path_template)

    # parse variables from path
    variables = re.findall(var_pattern, pathname)
    variables = variables[0] if isinstance(variables[0], tuple) else variables

    return dict(zip(var_names, variables))



from ast import arg
import importlib
import json
import os
import traceback
from typing import Any, Callable, Dict, List, Literal
from uuid import UUID, uuid4

from dash import html, dcc, Input, Output, State, MATCH, callback
from dash._get_paths import app_strip_relative_path
from dash._utils import inputs_to_vals
from dash._validate import validate_and_group_input_args
from dash.development.base_component import Component
from dash_router.core.context import RoutingContext
from dash import Dash
from dash._pages import _parse_query_string
from flask import request

from .utils.constants import DEFAULT_LAYOUT_TOKEN
from .utils.helper_functions import (
    format_relative_path,
    path_to_module,
    recursive_to_plotly_json,
    format_relative_path,
    _invoke_layout,
)
from .components import ChildContainer, LacyContainer, RootContainer, SlotContainer
from .core.routing import PageNode, RouteConfig, RouteTable, RouteTree, RouterResponse
from .core.execution import ExecNode
from .core.query_params import extract_function_inputs


class Router:
    def __init__(
        self,
        app: Dash,
        pages_folder: str = "pages",
        requests_pathname_prefix: str | None = None,
        ignore_empty_folders: bool = False,
    ) -> None:
        self.app = app
        self.static_routes = {}
        self.dynamic_routes = {}
        self.route_table = {}
        self.requests_pathname_prefix = requests_pathname_prefix
        self.ignore_empty_folders = ignore_empty_folders
        self.pages_folder = pages_folder

        if not isinstance(self.app, Dash):
            raise TypeError(f"App needs to be of Dash not: {type(self.app)}")

        self.setup_route_tree()
        self.setup_router()
        self.setup_lacy_callback()

    def setup_route_tree(self) -> None:
        """Sets up the route tree by traversing the pages folder."""
        root_dir = ".".join(self.app.server.name.split(os.sep)[:-1])
        self._traverse_directory(root_dir, self.pages_folder, None)

    def _validate_node(self, node: PageNode):
        # Validate Slots

        # Validate children
        if node.default_child:
            pass

    def _validate_tree(self):
        for root_node in self.dynamic_routes.routes.items():
            self._validate_node(root_node)

    def _traverse_directory(
        self,
        parent_dir: str,
        segment: str,
        current_node: PageNode | None,
    ) -> None:
        """Recursively traverses the directory structure and registers routes."""
        current_dir = os.path.join(parent_dir, segment)
        if not os.path.exists(current_dir):
            return

        entries = os.listdir(current_dir)
        dir_has_page = "page.py" in entries

        if dir_has_page:
            new_node = self.load_route_module(current_dir, segment, current_node)
            if new_node is not None:

                RouteTable.add_node(new_node)
                RouteTree.add_node(new_node, current_node)

                next_node = new_node
            else:
                next_node = current_node
        else:
            next_node = current_node

        for entry in sorted(entries):
            if entry.startswith((".", "_")) or entry == "page.py":
                continue

            full_path = os.path.join(current_dir, entry)
            if os.path.isdir(full_path):
                self._traverse_directory(current_dir, entry, next_node)

    def load_route_module(
        self, current_dir: str, segment: str, parent_node: PageNode
    ) -> PageNode | None:
        """Load modules and create Page Node"""
        relative_path = os.path.relpath(current_dir, self.pages_folder)
        relative_path = format_relative_path(relative_path)
        page_module_name = path_to_module(current_dir, "page.py")
        parent_node_id = parent_node.node_id if parent_node else None

        route_config = (
            self.import_route_component(current_dir, "page.py", "config")
            or RouteConfig()
        )
        is_static = route_config.is_static or relative_path == "/"
        is_root = parent_node and parent_node.segment == "/"
        segment = relative_path if is_static else segment

        page_layout = self.import_route_component(current_dir, "page.py")
        loading_layout = self.import_route_component(current_dir, "loading.py")
        error_layout = (
            self.import_route_component(current_dir, "error.py") or self.app._on_error
        )

        endpoint = self.import_route_component(current_dir, "api.py", "endpoint")
        endpoint_inputs, end_types = extract_function_inputs(endpoint)
        layout_inputs, inp_types = extract_function_inputs(page_layout)
        inputs = set(endpoint_inputs + layout_inputs)

        node_id = str(uuid4())
        new_node = PageNode(
            _segment=segment,
            node_id=node_id,
            layout=page_layout,
            parent_id=parent_node_id,
            module=page_module_name,
            is_root=is_root,
            error=error_layout,
            loading=loading_layout,
            endpoint=endpoint,
            endpoint_inputs=inputs,
            path=relative_path,
            is_static=is_static,
            default_child=route_config.default_child,
        )

        return new_node

    def strip_relative_path(self, path: str) -> str:
        return app_strip_relative_path(self.app.config.requests_pathname_prefix, path)

    def import_route_component(
        self,
        current_dir: str,
        file_name: Literal["page.py", "error.py", "loading.py", "api.py"],
        component_name: Literal["layout", "config", "endpoint"] = "layout",
    ) -> Callable[..., Component] | Component | None:
        page_module_name = path_to_module(current_dir, file_name)
        try:
            page_module = importlib.import_module(page_module_name)
            layout = getattr(page_module, component_name, None)
            if file_name == "page.py" and not layout:
                raise ImportError(
                    f"Module {page_module_name} needs a layout function or component"
                )
            return layout

        except ImportError as e:
            if file_name == "layout.py":
                print(f"Error processing {page_module_name}: {e}")
                print(f"Traceback: {traceback.format_exc()}")
                raise ImportError(
                    f"Module {page_module_name} needs a layout function or component"
                )
        except Exception as e:
            print(f"Error processing {page_module_name}: {e}")
            print(f"Traceback: {traceback.format_exc()}")

        return None

    def build_execution_tree(
        self,
        current_node: PageNode,
        ctx: RoutingContext,
    ) -> ExecNode:
        """
        Recursively builds the execution tree for the matched route.
        It extracts any path variables, processes child nodes, and handles slot nodes.
        """

        if not current_node:
            return current_node

        next_segment = ctx.peek_segment()
        segment_key = current_node.create_segment_key(next_segment)
        is_lacy = ctx.should_lazy_load(current_node, segment_key)

        if current_node.is_path_template:
            ctx.consume_path_var(current_node)
            next_segment = ctx.peek_segment()

        exec_node = ExecNode(
            segment=segment_key,
            node_id=current_node.node_id,
            layout=current_node.layout,
            parent_id=current_node.parent_id,
            variables=ctx.variables,
            loading=current_node.loading,
            error=current_node.error,
            is_lacy=is_lacy,
        )

        if is_lacy:
            ctx.set_node_state(current_node, "done", segment_key)
            return exec_node

        if current_node.endpoint and not DEFAULT_LAYOUT_TOKEN in segment_key:
            ctx.add_endpoint(current_node)

        if current_node.child_nodes or current_node.path_template:
            child_node = current_node.get_child_node(next_segment)

            if not child_node or not child_node.is_path_template:
                ctx.pop_segment()

            child_exec = self.build_execution_tree(
                current_node=child_node,
                ctx=ctx,
            )

            exec_node.child_node = child_exec

        if current_node.slots:
            exec_node.slots = self._process_slot_nodes(
                current_node=current_node,
                ctx=ctx,
            )

        ctx.set_node_state(current_node, "done", segment_key)
        return exec_node

    def _process_slot_nodes(
        self,
        current_node: PageNode,
        ctx: RoutingContext,
    ) -> Dict[str, ExecNode]:
        """Processes all slot nodes defined on the current node."""
        slot_exec_nodes: Dict[str, ExecNode] = {}
        for slot_name, slot_id in current_node.slots.items():
            slot_node = RouteTable.get_node(slot_id)
            segment_key = slot_node.create_segment_key(None)
            ctx.set_node_state(slot_node, "done", segment_key)

            slot_exec_node = self.build_execution_tree(
                current_node=slot_node,
                ctx=ctx,
            )

            slot_exec_nodes[slot_name] = slot_exec_node

        return slot_exec_nodes

    # ─── RESPONSE BUILDER ─────────────────────────────────────
    def resolve_url(
        self,
        pathname: str,
        query_parameters: Dict[str, Any],
        loading_state: Dict[str, Any],
    ) -> RouterResponse:

        path = self.strip_relative_path(pathname)
        ctx = RoutingContext.from_request(
            pathname=path,
            query_params=query_parameters,
            loading_state_dict=loading_state,
            resolve_type="url",
        )

        static_route, path_variables = RouteTree.get_static_route(ctx)
        if static_route:
            layout = _invoke_layout(
                static_route.layout, **query_parameters, **path_variables
            )
            return self.build_response(
                node=static_route, layout=layout, loading_state={}
            )

        active_node = RouteTree.get_active_root_node(ctx, self.ignore_empty_folders)

        if not active_node:
            return self.build_response(node=None, loading_state={})

        exec_tree = self.build_execution_tree(
            current_node=active_node,
            ctx=ctx,
        )

        if not exec_tree:
            return self.build_response(node=None, loading_state={})

        result_data = ctx.gather_endpoints()
        final_layout = exec_tree.execute(result_data)

        new_loading_state = {
            **ctx.get_updated_loading_state(),
            "query_params": query_parameters,
        }

        response = self.build_response(
            node=active_node, loading_state=new_loading_state, layout=final_layout
        )

        return response

    def resolve_search(
        self,
        pathname: str,
        query_params: Dict[str, any],
        updated_query_parameters: Dict[str, any],
        loading_state: Dict[str, any],
    ) -> RouterResponse:

        path = self.strip_relative_path(pathname)
        ctx = RoutingContext.from_request(
            pathname=path,
            loading_state_dict=loading_state,
            query_params=query_params,
            resolve_type="search",
        )

        # Collect all eligible nodes (nodes whose endpoint inputs match updated query parameters)
        eligible_nodes = []
        for _, loaded_node in ctx._loading_states.items():
            node_id = loaded_node.node_id
            node = RouteTable.get_node(node_id)
            if any(
                node_input in updated_query_parameters
                for node_input in node.endpoint_inputs
            ):
                eligible_nodes.append(node)

        if not eligible_nodes:
            return

        # Find parent nodes and create a set of nodes to process
        nodes_to_process_ids = set()
        for node in eligible_nodes:
            current = node
            while current:
                # If we find a parent that needs to be processed, add it and stop traversing
                if any(
                    parent_input in updated_query_parameters
                    for parent_input in current.endpoint_inputs
                ):
                    nodes_to_process_ids.add(current.node_id)
                    break
                current = (
                    RouteTable.get_node(current.parent_id)
                    if current.parent_id
                    else None
                )

        # If no parent nodes need processing, process the eligible nodes directly
        if not nodes_to_process_ids:
            nodes_to_process_ids = {node.node_id for node in eligible_nodes}

        # Build execution trees for all selected nodes
        exec_trees = []
        nodes_to_process = []
        for node_id in nodes_to_process_ids:
            node = RouteTable.get_node(node_id)
            exec_tree = self.build_execution_tree(
                current_node=node,
                ctx=ctx,
            )
            if exec_tree:
                exec_trees.append(exec_tree)
                nodes_to_process.append(node)

        # Gather all endpoints once
        endpoint_results = ctx.gather_endpoints()

        # Execute all trees with the same endpoint results
        layouts = []
        nodes = []
        for exec_tree, node in zip(exec_trees, nodes_to_process):
            layout = exec_tree.execute(endpoint_results)
            if layout:
                layouts.append(layout)
                nodes.append(node)

        new_loading_state = {
            **loading_state,
            **ctx.get_updated_loading_state(),
            "query_params": query_params,
        }

        return self.build_multi_response(nodes, new_loading_state, layouts)

    def build_response(
        self,
        node: PageNode,
        loading_state,
        layout: Component = None,
        remove_layout: bool = False,
    ):
        match node:
            case None:
                container_id = RootContainer.ids.container
                layout = html.H1("404 - Page not found")
                loading_state = {}

            case _ if node.is_root or node.is_static:
                container_id = (
                    RootContainer.ids.container
                    if not remove_layout
                    else json.dumps(ChildContainer.ids.container(node.node_id))
                )

            case _ if node.is_slot:
                container_id = json.dumps(
                    SlotContainer.ids.container(node.parent_id, node.segment)
                )

            case _:
                container_id = json.dumps(
                    ChildContainer.ids.container(
                        node.parent_id if not remove_layout else node.node_id
                    )
                )

        rendered_layout = recursive_to_plotly_json(layout)
        response = {
            container_id: {"children": rendered_layout},
            RootContainer.ids.state_store: {"data": loading_state},
        }
        return RouterResponse(multi=True, response=response)

    def build_multi_response(
        self, nodes: List[PageNode], loading_state, layouts: List[Component]
    ) -> RouterResponse:
        """Builds a response containing multiple layout updates with a single state store."""
        if not nodes or not layouts:
            return self.build_response(None, {})

        response = {}

        for node, layout in zip(nodes, layouts):
            single_response = self.build_response(node, loading_state, layout)
            response.update(single_response.response)

        return RouterResponse(multi=True, response=response)

    # ─── SYNCHRONOUS ROUTER SETUP ───────────────────────────────────────────────────
    def setup_router(self) -> None:
        @self.app.server.before_request
        def router():
            request_data = request.get_data()
            if not request_data:
                return

            body = json.loads(request_data)
            changed_prop = body.get("changedPropIds")

            if changed_prop:
                parts = changed_prop[0].split(".")
                changed_prop_id = parts[0]
                prop = parts[1] if len(parts) > 1 else None
            else:
                return

            if changed_prop_id != RootContainer.ids.location:
                return

            output = body["output"]
            inputs = body.get("inputs", [])
            state = body.get("state", [])
            cb_data = self.app.callback_map[output]
            inputs_state_indices = cb_data["inputs_state_indices"]
            print("inputs_state_indices: ", inputs_state_indices, flush=True)
            args = inputs_to_vals(inputs + state)
            print("ARGS: ", args, flush=True)
            pathname_, search_, loading_state_, states_ = args
            query_parameters = _parse_query_string(search_)
            previous_qp = loading_state_.pop("query_params", {})
            _, func_kwargs = validate_and_group_input_args(args, inputs_state_indices)
            func_kwargs = dict(list(func_kwargs.items())[3:])
            varibales = {**query_parameters, **func_kwargs}

            if prop == "pathname":
                try:
                    response = self.resolve_url(
                        pathname_, varibales, loading_state_
                    )
                    return response.model_dump()
                except Exception:
                    print(f"Traceback: {traceback.format_exc()}")
                    raise Exception("Failed to resolve the URL")

            if prop == "search":
                updated = dict(set(query_parameters.items()) - set(previous_qp.items()))
                missing_keys = previous_qp.keys() - query_parameters.keys()
                missing = {
                    key: None
                    for key in missing_keys
                    if key not in self.app.routing_callback_inputs
                }
                updates = dict(updated.items() | missing.items())
                response = self.resolve_search(
                    pathname_, varibales, updates, loading_state_
                )
                return response.model_dump() if response else response

        with self.app.server.app_context():
            inputs = dict(
                pathname_=Input(RootContainer.ids.location, "pathname"),
                search_=Input(RootContainer.ids.location, "search"),
                loading_state_=State(RootContainer.ids.state_store, "data"),
            )
            inputs.update(self.app.routing_callback_inputs)

            @self.app.callback(
                Output(RootContainer.ids.dummy, "children"),
                inputs=inputs,
            )
            def update(
                pathname_: str, search_: str, loading_state_: str, **states
            ):
                pass

    def setup_lacy_callback(self):
        inputs = dict(
            lacy_segment_id=Input(LacyContainer.ids.container(MATCH), "id"),
            variables=Input(LacyContainer.ids.container(MATCH), "data-path"),
            pathname=State(RootContainer.ids.location, "pathname"),
            search=State(RootContainer.ids.location, "search"),
            loading_state=State(RootContainer.ids.state_store, "data"),
        )

        @self.app.callback(
            Output(LacyContainer.ids.container(MATCH), "children"), inputs=inputs
        )
        def load_lacy_component(
            lacy_segment_id, variables, pathname, search, loading_state
        ):
            node_id = lacy_segment_id.get("index")
            qs = _parse_query_string(search)
            query_parameters = loading_state.pop("query_params", {})
            node_variables = json.loads(variables)
            variables = {**qs, **query_parameters, **node_variables}

            lacy_node = RouteTable.get_node(node_id)
            path = self.strip_relative_path(pathname)
            segments = path.split("/")
            node_segments = lacy_node.module.split(".")[1:-1]
            current_index = node_segments.index(lacy_node.segment_value)
            remaining_segments = list(reversed(segments[current_index:]))

            ctx = RoutingContext.from_request(
                pathname=pathname,
                query_params=variables,
                loading_state_dict=loading_state,
                resolve_type="lacy",
            )

            ctx.segments = remaining_segments

            exec_tree = self.build_execution_tree(
                current_node=lacy_node,
                ctx=ctx,
            )

            endpoint_results = ctx.gather_endpoints()
            layout = exec_tree.execute(endpoint_results)

            return layout
